/*
 * colbyhpdl100a device support
 */

#include <epicsStdio.h>
#include <devCommonGpib.h>
#include <math.h>

/******************************************************************************
 *
 * The following define statements are used to declare the names to be used
 * for the dset tables.   
 *
 * A DSET_AI entry must be declared here and referenced in an application
 * database description file even if the device provides no AI records.
 *
 ******************************************************************************/
#define DSET_AI     devAicolbyhpdl100a
#define DSET_AO     devAocolbyhpdl100a
#define DSET_BI     devBicolbyhpdl100a
#define DSET_BO     devBocolbyhpdl100a
#define DSET_EV     devEvcolbyhpdl100a
#define DSET_LI     devLicolbyhpdl100a
#define DSET_LO     devLocolbyhpdl100a
#define DSET_MBBI   devMbbicolbyhpdl100a
#define DSET_MBBID  devMbbidcolbyhpdl100a
#define DSET_MBBO   devMbbocolbyhpdl100a
#define DSET_MBBOD  devMbbodcolbyhpdl100a
#define DSET_SI     devSicolbyhpdl100a
#define DSET_SO     devSocolbyhpdl100a
#define DSET_WF     devWfcolbyhpdl100a

#include <devGpib.h> /* must be included after DSET defines */

#define TIMEOUT     1.0    /* I/O must complete within this time */
#define TIMEWINDOW  2.0    /* Wait this long after device timeout */

static char *channelNames[] = { "", "1", "2" };

/*
 * Custom conversion routine for 'set delay' command
 * Converts value from nanoseconds to seconds and does the rounding
 * and special (annoying) formatting that this device requires.
 */
static int
convertSetDelay(gpibDpvt *pdpvt, int P1, int P2, char **P3)
{
    struct aoRecord *pao = (struct aoRecord *)pdpvt->precord;
    double rounded = floor(pao->val * 100 + 0.5) / 100.0;

    sprintf(pdpvt->msg, "DEL%s %.4E\n", P3[P1], rounded * 1.0e-9);
    return 0;
}

/*
 * Custom conversion routine for 'get delay' command
 * Converts value from seconds to nanoseconds
 */
static int
convertGetDelay(gpibDpvt *pdpvt, int P1, int P2, char **P3)
{
    struct aiRecord *pai = (struct aiRecord *)pdpvt->precord;
    double d;

    if (sscanf(pdpvt->msg, "%lf", &d) != 1) {
        epicsSnprintf(pdpvt->pasynUser->errorMessage,
                      pdpvt->pasynUser->errorMessageSize,
                     "scanf failed");
        return -1;
    }
    pai->val = d * 1.0e9;
    pai->udf = 0;
    return 0;
}

/******************************************************************************
 * Strings used by the init routines to fill in the znam,onam,...
 * fields in BI and BO record types.
 ******************************************************************************/
static char  *resetList[] = { "Reset","Reset" };
static struct devGpibNames reset = { 2,resetList,0,1 };

static char  *clearList[] = { "Clear","Clear" };
static struct devGpibNames clear = { 2,clearList,0,1 };

/******************************************************************************
 * Array of structures that define all GPIB messages
 * supported for this type of instrument.
 ******************************************************************************/
static struct gpibCmd gpibCmds[] = {
    /* Param 0 -- Read SCPI identification string */
    {&DSET_SI, GPIBREAD, IB_Q_HIGH, "*IDN?\n", "%39[^\r\n]", 0, 200, NULL, 0, 0, NULL, NULL, "\n" },

    /* Param 1 - SCPI reset command */
    {&DSET_BO, GPIBCMD, IB_Q_HIGH, "*RST\n", NULL, 0, 80, NULL, 0, 0, NULL, &reset, "\n" },

    /* Param 2 - SCPI clear status command */
    {&DSET_BO, GPIBCMD, IB_Q_HIGH, "*CLS\n", NULL, 0, 80, NULL, 0, 0, NULL, &clear, "\n" },

    /* Param 3 - Read SCPI status byte */
    {&DSET_LI, GPIBREAD, IB_Q_HIGH, "*STB?\n", "%d", 0, 80, NULL, 0, 0, NULL, NULL, "\n" },

    /* Param 4 - Read SCPI event register */
    {&DSET_LI, GPIBREAD, IB_Q_HIGH, "*ESR?\n", "%d", 0, 80, NULL, 0, 0, NULL, NULL, "\n" },

    /* Param 5 - Enable SCPI events */
    {&DSET_LO, GPIBWRITE, IB_Q_HIGH, NULL, "*ESE %d\n", 0, 80, NULL, 0, 0, NULL, NULL, "\n" },

    /* Param 6 - Read back SCPI enabled events */
    {&DSET_LI, GPIBREAD, IB_Q_HIGH, "*ESE?\n", "%d", 0, 80, NULL, 0, 0, NULL, NULL, "\n" },

    /* Param 7 - Enable SCPI service request sources */
    FILL,

    /* Param 8 - Read back SCPI enabled service request sources */
    FILL,

    /* Param 9 - Read SCPI output completion status */
    {&DSET_LI, GPIBREAD, IB_Q_HIGH, "*OPC?\n", "%d", 0, 80, NULL, 0, 0, NULL, NULL, "\n" },

    /* Param 10 - Read delay */
    {&DSET_AI, GPIBREAD, IB_Q_HIGH, "DEL?\n", NULL, 0, 80, convertGetDelay, 0, 0, NULL, NULL, "\n" },

    /* Param 11 - Write delay(s) */
    {&DSET_AO, GPIBWRITE, IB_Q_HIGH, NULL, NULL, 0, 80, convertSetDelay, 0, 0, channelNames, NULL, "\n" },

    /* Param 12 - Read channel 1 delay */
    {&DSET_AI, GPIBREAD, IB_Q_HIGH, "DEL1?\n", NULL, 0, 80, convertGetDelay, 0, 0, NULL, NULL, "\n" },

    /* Param 13 - Write channel 1 delay */
    {&DSET_AO, GPIBWRITE, IB_Q_HIGH, NULL, NULL, 0, 80, convertSetDelay, 1, 0, channelNames, NULL, "\n" },

    /* Param 14 - Read channel 2 delay */
    {&DSET_AI, GPIBREAD, IB_Q_HIGH, "DEL2?\n", NULL, 0, 80, convertGetDelay, 0, 0, NULL, NULL, "\n" },

    /* Param 15 - Write channel 2 delay */
    {&DSET_AO, GPIBWRITE, IB_Q_HIGH, NULL, NULL, 0, 80, convertSetDelay, 2, 0, channelNames, NULL, "\n" },
};

/* The following is the number of elements in the command array above.  */
#define NUMPARAMS sizeof(gpibCmds)/sizeof(struct gpibCmd)

/******************************************************************************
 * Initialize device support parameters
 *
 *****************************************************************************/
static long init_ai(int parm)
{
    if(parm==0) {
        devSupParms.name = "devcolbyhpdl100a";
        devSupParms.gpibCmds = gpibCmds;
        devSupParms.numparams = NUMPARAMS;
        devSupParms.timeout = TIMEOUT;
        devSupParms.timeWindow = TIMEWINDOW;
        devSupParms.respond2Writes = -1;
    }
    return(0);
}
